import re
from functools import singledispatch
from typing import Union

from ruamel.yaml.comments import CommentedMap, CommentedSeq

from souschef.config import RecipeConfiguration


@singledispatch
def convert_to_abstract_repr(
    item: Union[None, CommentedSeq, CommentedMap],
    name: str,
    yaml,
    config: RecipeConfiguration,
):
    return item


@convert_to_abstract_repr.register
def convert_to_abstract_repr_commented_map(
    item: CommentedMap, name: str, yaml, config: RecipeConfiguration
):
    from souschef.section import Section

    return Section(name=name, item=item, parent=yaml, config=config)


@convert_to_abstract_repr.register
def convert_to_abstract_repr_commented_seq(
    item: CommentedSeq, name: str, yaml, config: RecipeConfiguration
):
    from souschef.ingredient import IngredientList

    return IngredientList(parent_yaml=item, key=name, config=config)


@convert_to_abstract_repr.register
def convert_to_abstract_repr_str(
    item: str, name: str, yaml, config: RecipeConfiguration
):
    from souschef.ingredient import Ingredient

    return Ingredient(parent=yaml, position=name)


@convert_to_abstract_repr.register
def convert_to_abstract_repr_int(
    item: int, name: str, yaml, config: RecipeConfiguration
):
    from souschef.ingredient import Ingredient

    return Ingredient(parent=yaml, position=name)


def parse_value(value):
    selector = ""
    if isinstance(value, str) and "#" in value:
        re_match = re.match(r"(.*?)\s*#(.*)", value)
        value, selector = re_match.groups()
    return value, selector



import re
from typing import Union

from souschef.comment import Comment
from souschef.recipe import Recipe

RE_CHECK_JINJA = re.compile(r"\s*([#{]\s*%.*?\s*%\s*}?)")
RE_JINJA_SET_EXPRESSION = re.compile(
    r"[#{]\s*%\s+set\s+(\w+)\s*=\s*[\'\"]*(.*?)[\'\"]*\s*%\s*}",
    re.DOTALL | re.IGNORECASE,
)


def is_jinja_expression(recipe_item):
    if isinstance(recipe_item, Comment):
        return RE_CHECK_JINJA.search(recipe_item.raw_value) is not None
    return False


def get_all_jinja_expression(recipe: Recipe):
    for item in recipe:
        if is_jinja_expression(item):
            yield item


def set_global_jinja_var(recipe: Recipe, var_name: str, new_value: Union[str, int]):
    all_jinja_exp = list(get_all_jinja_expression(recipe))
    if not all_jinja_exp:
        recipe.yaml.yaml_set_start_comment("NEW VALUE")
        recipe.yaml.ca.comment[1][0].value = f'#% set {var_name} = "{new_value}" %}}\n'
        return
    for item in all_jinja_exp:
        all_jinja = RE_CHECK_JINJA.split(item.raw_value)
        result = []
        found_var = False
        for single_jinja in all_jinja:
            match_jinja = RE_JINJA_SET_EXPRESSION.match(single_jinja)
            if match_jinja:
                current_var, current_value = match_jinja.groups()
                if current_var == var_name:
                    found_var = True
                    single_jinja = f'#% set {current_var} = "{new_value}" %}}'
            result.append(single_jinja)
        if found_var:
            result = "".join(result)
            result = result.replace("}}#%", "}}\n#%")
            item.raw_value = result.strip()
            return
    all_jinja_exp[-1].raw_value += f'\n#% set {var_name} = "{new_value}" %}}'


def get_global_jinja_var(recipe: Recipe, var_name: str) -> str:
    for item in recipe:
        if not is_jinja_expression(item):
            continue
        all_jinja = RE_CHECK_JINJA.split(item.raw_value)
        for single_jinja in all_jinja:
            match_jinja = RE_JINJA_SET_EXPRESSION.match(single_jinja)
            if match_jinja:
                current_var, current_value = match_jinja.groups()
                if current_var == var_name:
                    return current_value
    raise ValueError(
        f"It was not possible to find the requested jinja variable '{var_name}'."
    )



import re
from typing import List

from ruamel.yaml import CommentToken

from souschef.config import NEW_LINE


class Comment:
    def __init__(self, yaml, pos: int = 0):
        self._yaml = yaml
        self.__pos = pos

    @property
    def raw_value(self) -> str:
        return self._yaml.value.split(NEW_LINE)[self.__pos]

    @raw_value.setter
    def raw_value(self, val: str):
        self.value = val

    @property
    def value(self) -> str:
        all_values = self._yaml.value.split(NEW_LINE)
        if not all_values:
            return ""
        re_comment = re.search(r"^\s*#\s*(.*)", all_values[self.__pos], re.DOTALL)
        if re_comment is None:
            return all_values[self.__pos]
        try:
            return re_comment.group(1)
        except IndexError:
            return all_values[self.__pos]

    @value.setter
    def value(self, new_value: str):
        re_comment = re.compile(r"^(\s*#\s*)(.*)", re.DOTALL)
        if re_comment.search(new_value) is None:
            new_value = f"# {new_value}"

        all_values = self._yaml.value.split(NEW_LINE)
        all_values = all_values[:-1] if not all_values[-1] else all_values
        all_values[self.__pos] = new_value
        self._yaml.value = f"{NEW_LINE.join(all_values)}\n"

    def remove(self):
        all_values = self._yaml.value.split(NEW_LINE)
        del all_values[self.__pos]
        self._yaml.value = f"{NEW_LINE.join(all_values)}"

    def __repr__(self) -> str:
        return repr(self.raw_value)

    def __str__(self) -> str:
        return str(self.value)

    def __len__(self) -> int:
        return len(str(self))

    def __eq__(self, other: str) -> bool:
        separated_comment = re.search(r"^\s*#\s*(.*)", other)
        if separated_comment is None:
            return self.value == other or self.raw_value == other
        return self.value == separated_comment.group(1)


def comment_factory(
    comment_token: CommentToken, start_inline_comment: bool = False
) -> List[Comment]:
    if comment_token is None:
        return []
    num_lines = max(0, len(comment_token.value.split("\n")) - 1)
    start_pos = 1 if start_inline_comment else 0
    return [Comment(comment_token, pos) for pos in range(start_pos, num_lines)]



import weakref
from collections import abc
from typing import Any, Union

from ruamel.yaml.comments import CommentedMap, CommentedSeq

from souschef import mixins
from souschef.tools import convert_to_abstract_repr, parse_value


class Section(
    mixins.SelectorMixin,
    mixins.GetSetItemMixin,
    mixins.InlineCommentMixin,
    mixins.AddSection,
):
    def __init__(
        self,
        name: Union[int, str],
        item: CommentedMap,
        parent: CommentedMap,
        config,
    ):
        self.__yaml = weakref.ref(item)
        self._parent = weakref.ref(parent)
        self._config = weakref.ref(config)
        self._name = name

    def __repr__(self) -> str:
        return f"<Section {self._name}>"

    def items(self):
        for k, v in self.yaml.items():
            yield k, convert_to_abstract_repr(v, k, self._parent(), self._config())

    def keys(self):
        return self.yaml.keys()

    def values(self):
        for _, val in self.items():
            yield val

    @property
    def yaml(self):
        return self.__yaml()

    def __str__(self) -> str:
        return self._name

    def __contains__(self, item):
        return item in self.yaml

    @property
    def value(self):
        return [v for v in self]

    @value.setter
    def value(self, items: Any):
        if isinstance(items, abc.Mapping):
            for key, value in items.items():
                self[key] = value
        elif isinstance(items, str):
            pkg_info, comment = parse_value(items)
            self._parent()[self._name] = pkg_info
            self.inline_comment = comment
        elif isinstance(items, abc.Sequence):
            self._parent()[self._name] = CommentedSeq()
            for pos, i in enumerate(items):
                pkg_info, comment = parse_value(i)
                self._parent()[self._name].append(pkg_info)
                if comment:
                    self[pos].inline_comment = comment
        else:
            self._parent()[self._name] = items

    def update(self, section: Any):
        self.value = section
